# ----------------------------------------------------------------------------------------------------
# IBM Confidential
# Licensed Materials - Property of IBM
# 5737-H76, 5900-A3Q
# © Copyright IBM Corp. 2025  All Rights Reserved.
# US Government Users Restricted Rights - Use, duplication or disclosure restricted by
# GSA ADPSchedule Contract with IBM Corp.
# ----------------------------------------------------------------------------------------------------

from typing import Annotated

from pydantic import BaseModel, Field, model_validator
from typing_extensions import Self

from ibm_watsonx_gov.entities.credentials import (AzureOpenAICredentials,
                                                  OpenAICredentials,
                                                  RITSCredentials,
                                                  WxAICredentials,
                                                  )
from ibm_watsonx_gov.entities.enums import ModelProviderType


class ModelProvider(BaseModel):
    type: Annotated[
        ModelProviderType, Field(
            description="The type of model provider.")
    ]


class WxAIModelProvider(ModelProvider):
    type: Annotated[ModelProviderType, Field(
        description="The type of model provider.", default=ModelProviderType.IBM_WATSONX_AI)]
    credentials: Annotated[WxAICredentials | None, Field(default=None)]

    @model_validator(mode="after")
    def create_credentials_from_env(self) -> Self:
        if self.credentials is None:
            try:
                self.credentials = WxAICredentials.create_from_env()
            except ValueError:
                self.credentials = None
        return self


class OpenAIModelProvider(ModelProvider):
    type: Annotated[ModelProviderType, Field(
        description="The type of model provider.", default=ModelProviderType.OPENAI)]
    credentials: Annotated[OpenAICredentials | None, Field(
        description="OpenAI credentials. This can also be set by using `OPENAI_API_KEY` environment variable.", default=None)]

    @model_validator(mode="after")
    def create_credentials_from_env(self) -> Self:
        if self.credentials is None:
            self.credentials = OpenAICredentials.create_from_env()
        return self


class AzureOpenAIModelProvider(ModelProvider):
    type: Annotated[ModelProviderType, Field(
        description="The type of model provider.", default=ModelProviderType.AZURE_OPENAI)]
    credentials: Annotated[AzureOpenAICredentials | None, Field(
        description="Azure OpenAI credentials.", default=None
    )]

    @model_validator(mode="after")
    def create_credentials_from_env(self) -> Self:
        if self.credentials is None:
            self.credentials = AzureOpenAICredentials.create_from_env()
        return self


class RITSModelProvider(ModelProvider):
    type: Annotated[ModelProviderType, Field(
        description="The type of model provider.", default=ModelProviderType.RITS)]
    credentials: Annotated[RITSCredentials | None, Field(
        description="RITS credentials.", default=None
    )]

    @model_validator(mode="after")
    def create_credentials_from_env(self) -> Self:
        if self.credentials is None:
            self.credentials = RITSCredentials.create_from_env()
        return self


class CustomModelProvider(ModelProvider):
    type: Annotated[ModelProviderType, Field(
        description="The type of model provider.", default=ModelProviderType.CUSTOM)]
